package ssw.mj.test;

import java.io.StringReader;

import ssw.mj.Parser;
import ssw.mj.Scanner;
import ssw.mj.codegen.Code;
import ssw.mj.symtab.Obj;
import ssw.mj.symtab.Struct;
import ssw.mj.symtab.Tab;

/**
 * Test cases for the examples from the lab sessions 8 and 9.
 * 
 * @author AW
 */
public class UEBeispieleTest extends IdentifiedTestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";
	
	private static TestPrintWriter output;
	private static TestByteArrayOutputStream code;

	// ---------- helpers ----------
	
	private void initParser (String program) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(program), null);
		code = new TestByteArrayOutputStream();
	}

	public void testBspSymTab () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		////////////////////////		
		// check symbol table //
		////////////////////////
		Obj actual, expected, locals;
		Struct bT;

		actual = Tab.find("A");
		expected = new Obj(Obj.Prog, "A", Tab.noType);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'A'", expected, actual);
		
		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages
		
		actual = actual.locals;
		expected = new Obj(Obj.Con, "max", Tab.intType);
		expected.val = 12;
		assertEquals("Constant 'max'", expected, actual);
				
		actual = actual.next;
		expected = new Obj(Obj.Var, "c", Tab.charType);
		expected.adr = 0;
		expected.level = 0;
		assertEquals("Global 'c'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Var, "i", Tab.intType);
		expected.adr = 1;
		expected.level = 0;
		assertEquals("Variable 'i'", expected, actual);

		actual = actual.next;
		bT = new Struct(Struct.Class);
		bT.n = 2;
		expected = new Obj(Obj.Var, "x", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Field 'x'", expected, actual.type.fields);
		bT.fields = expected;
		expected = new Obj(Obj.Var, "y", Tab.intType);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Field 'y'", expected, actual.type.fields.next);
		bT.fields.next = expected;
		expected = new Obj(Obj.Type, "B", bT);
		assertEquals("Class 'B'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Var, "iarr", new Struct(Struct.Arr, Tab.intType));
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Local 'iarr'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Var, "b", bT);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Local 'b'", expected, actual.locals.next);
		locals.next = expected;
		expected = new Obj(Obj.Var, "n", Tab.intType);
		expected.adr = 2;
		expected.level = 1;
		assertEquals("Local 'n'", expected, actual.locals.next.next);
		locals.next.next = expected;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = locals;
		assertEquals("Methode 'main'", expected, actual);
	}

	public void testBsp01 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    n = 3;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			//   n = 3;
		code.expect(18, 1);   // const_3
		code.expect(9, 1);    // store_2

		Code.write(code);

		code.setCodeSize(2);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}
	
	public void testBsp02 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    i = 10;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			//   i = 10;
		code.expect(22, 1);   // const
		code.expect(10, 4);   // w
		code.expect(12, 1);   // putstatic
		code.expect(1, 2);    // s

		Code.write(code);

		code.setCodeSize(8);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}
	
	public void testBsp03 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    n = 3 + i;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			//   n = 3 + i;
		code.expect(18, 1);   // const_3
		code.expect(11, 1);   // getstatic
		code.expect(1, 2);    // s
		code.expect(23, 1);   // add
		code.expect(9, 1);    // store_2

		Code.write(code);

		code.setCodeSize(6);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}
	
	public void testBsp04 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    n = 3 + i * max - n;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			//   n = 3 + i * max - n;
		code.expect(18, 1);   // const_3
		code.expect(11, 1);   // getstatic
		code.expect(1, 2);    // s
		code.expect(22, 1);   // const
		code.expect(12, 4);   // w
		code.expect(25, 1);   // mul
		code.expect(23, 1);   // add
		code.expect(4, 1);    // load_2
		code.expect(24, 1);   // sub
		code.expect(9, 1);    // store_2

		Code.write(code);

		code.setCodeSize(14);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}

	public void testBsp05 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    iarr = new int[10];" + LF +
				"    iarr[5] = 10;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
		  //   iarr = new int[10];
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(33, 1);  // newarray
		code.expect( 1, 1);  // b
		code.expect( 7, 1);  // store_0
			//   iarr[5] = 10;
		code.expect( 2, 1);  // load_0
		code.expect(20, 1);  // const_5
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(35, 1);  // astore

		Code.write(code);

		code.setCodeSize(16);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}
	
	public void testBsp06 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    iarr = new int[10];" + LF +
				"    iarr[5] = 10;" + LF +
				"    b = new B;" + LF +
				"    b.y = iarr[5] * 3;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
		  //   iarr = new int[10];
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(33, 1);  // newarray
		code.expect( 1, 1);  // b
		code.expect( 7, 1);  // store_0
			//   iarr[5] = 10;
		code.expect( 2, 1);  // load_0
		code.expect(20, 1);  // const_5
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(35, 1);  // astore
			//   b = new B;
		code.expect(32, 1);  // new
		code.expect( 8, 2);  // s
		code.expect( 8, 1);  // store_1
			//   b.y = iarr[5] * 3;
		code.expect( 3, 1);  // load_1
		code.expect( 2, 1);  // load_0
		code.expect(20, 1);  // const_5
		code.expect(34, 1);  // aload
		code.expect(18, 1);  // const_3
		code.expect(25, 1);  // mul
		code.expect(14, 1);  // putfield
		code.expect( 1, 2);  // s

		Code.write(code);

		code.setCodeSize(29);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}

	public void testBsp07 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    n--;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
		  //   n--;
		code.expect(31, 1);  // inc
		code.expect( 2, 1);  // b1
		code.expect(-1, 1);  // b2

		Code.write(code);

		code.setCodeSize(3);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}

	public void testBsp08 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    i--;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
		  //   i--;
		code.expect(11, 1);  // getstatic
		code.expect( 1, 2);  // s
		code.expect(21, 1);  // const_m1
		code.expect(23, 1);  // add
		code.expect(12, 1);  // putstatic
		code.expect( 1, 2);  // s

		Code.write(code);

		code.setCodeSize(8);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}

	public void testBsp09 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    b = new B;" + LF +
				"    b.y--;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
		  //   b = new B;
		code.expect(32, 1);  // new
		code.expect( 8, 2);  // s
		code.expect( 8, 1);  // store_1
		  //   b.y--;
		code.expect( 3, 1);  // load_1
		code.expect(40, 1);  // dup
		code.expect(13, 1);  // getfield
		code.expect( 1, 2);  // s
		code.expect(21, 1);  // const_m1
		code.expect(23, 1);  // add
		code.expect(14, 1);  // putfield
		code.expect( 1, 2);  // s

		Code.write(code);

		code.setCodeSize(14);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}
	
	public void testBsp10 () {
		initParser(
				"program A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    iarr = new int[10];" + LF +
				"    iarr[0]--;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
		  //   iarr = new int[10];
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(33, 1);  // newarray
		code.expect( 1, 1);  // b
		code.expect( 7, 1);  // store_0
			//   iarr[0]--;
		code.expect( 2, 1);  // load_0
		code.expect(15, 1);  // const_0
		code.expect(41, 1);  // dup2
		code.expect(34, 1);  // aload
		code.expect(21, 1);  // const_m1
		code.expect(23, 1);  // add
		code.expect(35, 1);  // astore

		Code.write(code);

		code.setCodeSize(15);
		code.setDataSize(2);
		code.setMainPC(-1);

		code.verify();			
	}

	public static void main (String[] args) {
		junit.textui.TestRunner.run(UEBeispieleTest.class);
	}
}
